
#include "..\..\common\ImageFilter.h"
#include <string.h>
#include <math.h>
#include <vector>
#include <algorithm>

#pragma warning (disable:4800)

using namespace std;
using namespace FilterSpace;
using namespace System::Windows::Forms;


//////////////////////////////////
// FILTER-CORE-MACRO  <-- BEGIN //
//////////////////////////////////

#define FCORE(XPARAM, YPARAM) \
{ \
	int z = sx + x;	\
	bool apply = !mask;	\
	if (mask) apply = mask[z >> 3] & ((z & 0x07) ? (1 << (z & 0x07)) : 1); \
	\
	if (apply) \
	{ \
		for (int ch = 0; ch < channels; ++ch) \
		{ \
			int kernel_buffer_index = 0; \
			int x_in = x + x_pre; \
			\
			for (int my = -radius; my <= radius; my++) \
			{ \
				byte* line = virtual_image[ (YPARAM) ]; \
				int   from = MAX(x_in - radius, 0);	\
				int   to   = MIN(x_in + radius, x_caching - 1);	\
				\
				for (int mx = 0; mx <= radius * 2; ++mx) \
					kernel_array[kernel_buffer_index++] = line[ (XPARAM) * channels + ch ]; \
			} \
			nth_element(kernel_array.begin(), kernel_array.begin() + 4, kernel_array.end()); \
			*dout++ = kernel_array[4]; \
		} \
	} \
	else \
	{ \
		for (int ch = 0; ch < channels; ++ch) *dout++ = virtual_image[y_in][(x + sx) * channels + ch]; \
	} \
}
////////////////////////////////
// FILTER-CORE-MACRO  <-- END //
////////////////////////////////


///////////////////////
//       FILTER      //
///////////////////////

bool FilterCore(filtersettings* f, int sx, int sy, int sw, int sh, Semaphore^ updater)
{
	if      (!f)           return false;
	else if (!(f->datain)) return false;

	#define radius  1   // SET IT TO THE CORRESPONDING RADIUS VALUE == f->nCoreCriticalOverlap
	#define radius2 2

	// BEGIN --> COMMON SUPPORT CODE (DO NOT MODIFY!)

	// ImageFilterAccelerator RANGECHECKS sx, sy, sw, sh AND SET THEM CORRECTLY
	// IF YOU DO NOT USE ImageFilterAccelerator, YOU HAVE TO DO IT HERE MANUALLY 
	ImageFilterAccelerator* img = new ImageFilterAccelerator(f, sx, sy, sw, sh, radius);
	byte** virtual_image = img->GetVirtualImage();

	vector<int> kernel_array;
	kernel_array.resize(9);

	int x_pre  = img->GetPreX();
	int x_post = img->GetPostX();
	int y_pre  = img->GetPreY();
	int y_post = img->GetPostY();	
	int x_caching = sw + x_pre + x_post;
	int y_caching = sh + y_pre + y_post;

	int sout = f->strideout;
	int mstride = (f->width + 7) >> 3;
	int channels = f->channelcountin;
	byte *mask = f->mask;
	byte *dout = 0;

	// END --> COMMON SUPPORT CODE (DO NOT MODIFY!)

	int width = f->width;
	int height = f->height;
	int x, y, processed_pixels = 0;

	for (y = 0; y < sh; y++)
	{
		int y_in  = y + y_pre;
		dout = f->dataout + sout * y;

		img->StepImageScanLine();

		if (y < radius - y_pre || y >= sh + y_post - radius) // TOP AND BOTTOM "INCOMPLETE" ROWS
		{
			for (x = 0; x < sw; x++) FCORE(INTERVAL_LIMITER(from + mx, 0, x_caching - 1), INTERVAL_LIMITER(y_in + my, 0, y_caching - 1))
		}
		else
		{
			// LEFT "INCOMPLETE" COLUMNS
			for (x = 0; x < radius - x_pre; x++)					FCORE(INTERVAL_LIMITER(from + mx, 0, x_caching - 1), y_in + my)

			// COMPLETE CENTER REGION
			for (x = radius - x_pre; x < sw + x_post - radius; x++) FCORE(from + mx, y_in + my)

			// RIGHT "INCOMPLETE" COLUMNS
			for (x = sw + x_post - radius; x < sw; x++)				FCORE(INTERVAL_LIMITER(from + mx, 0, x_caching - 1), y_in + my)
		}

		// FOR PERFORMANCE MEASUREMENT
		processed_pixels += sw;

		// CONDITION CAN BE COMBINED WITH E.G. " && (processed_pixels > 1000)" 
		// TO AVOID FREQUENT MUTEX LOCKS

		if (updater != nullptr)
		{
			if (updater->WaitOne(10, false))
			{
				*(f->processing_ptr) += processed_pixels;
				processed_pixels = 0;
				updater->Release(1);
			}
		}
		
		else
		{
			if (f->processing_ptr) *(f->processing_ptr) += processed_pixels;
			processed_pixels = 0;
		}
		
		if (f->nCoreKillSignal) { if (*(f->nCoreKillSignal)) break; }
	}

	delete img;

	// FILTER CODE END -->

	return true;
}




IntPtr ImageFilter::CreateFilterSettings(Byte* fs, bool silent)
{
	int indata = 0, outdata = 0, stride = 0, height = 0, run_local = 1;

	if (!fs) return IntPtr(0);

	xml_node^ xn = gcnew xml_node( fs );

	filtersettings* f = XmlToMyStruct( xn, true );

	// GET "Run_Local" PARAMETER
	xn->XmlGetChildValueByName("Run_Local", run_local);

	// CLONE "InputImage0" TO "OutputImage0"
	xml_node^ xt = nullptr;

	if (xn->XmlGetChildByName(xt, "InputImage0" )) xt->XmlGetChildValueByName("Data", indata);
	if (xn->XmlGetChildByName(xt, "OutputImage0")) xt->XmlGetChildValueByName("Data", outdata);

	xn->XmlRemoveChildByName("OutputImage0");
	xn->XmlGetChildByName(xt, "InputImage0");

	Byte* fb = xt->XML_Tree_Marshal();
	xt = gcnew xml_node(fb);
	delete [] fb;
	xt->name = "OutputImage0";

	// SET "Run_Local" settings
	if (!run_local)
	{
		if (f->dataout == 0 || f->datain == f->dataout)  // FIRST RUN OF CreateFilterSettings
		{
			stride = f->stridein;
			height = f->height;
			if (stride * height == 0) return IntPtr(0);
			xt->XmlSetChildValueByName("Data", ((int)(void*)(new Byte[stride * height])));
		}
		else
		{
			xt->XmlSetChildValueByName("Data", outdata);
		}
	}
	xn->children.Add( xt );

	xn->XmlSetChildValueByName("nCore!ExecutionSize", 1); // RADIUS == 1 ALWAYS FOR THIS FILTER

	delete f;
	Byte* res = xn->XML_Tree_Marshal();
	delete xn;
	return (IntPtr)(void*)res;
}


filtersettings* ImageFilter::XmlToMyStruct(xml_node^ xn, bool creating)
{
	xml_node^ xntmp;

	if (! xn->XmlGetChildByName(xntmp, "InputImage0")) return 0;
	filtersettings* filset = new filtersettings;
	int packed, interleaved;
	if (GetOneImageSetting(xntmp, filset->width, filset->height, filset->stridein, packed, interleaved, filset->channelcountin, filset->pixelsizein, filset->datain)) 
	{
		MessageBox::Show("XmlToMyStruct - 0 - " + creating.ToString());
		delete filset;
		return 0;
	}

	if (! xn->XmlGetChildByName(xntmp, "OutputImage0") && !creating)
	{
		MessageBox::Show("XmlToMyStruct - 1 - " + creating.ToString());
		delete filset;
		return 0;
	}

	if (GetOneImageSetting(xntmp, filset->width, filset->height, filset->strideout, packed, interleaved, filset->channelcountout, filset->pixelsizeout, filset->dataout))
	{
		MessageBox::Show("XmlToMyStruct - 2 - " + creating.ToString());
		delete filset;
		return 0;
	}

	int m = 0;
	if ( xn->XmlGetChildByName(xntmp, "Mask") )	xntmp->XmlGetChildValueByName("Data", m);
	filset->mask = (Byte*)(void*)m;

	filset->processing_size    = ( xn->XmlGetChildValueByName("Processing_Size", m) )  ? (int*)(void*)m : 0;
	filset->processing_ptr     = ( xn->XmlGetChildValueByName("Processing_Ptr", m) )   ? (int*)(void*)m : 0;
	filset->nCoreEnabled       = ( xn->XmlGetChildValueByName("nCore!Enabled", m) )    ? m : 0;
	filset->nCoreKillSignal    = ( xn->XmlGetChildValueByName("nCore!KillSignal", m) ) ? (int*)(void*)m : 0;
	filset->nCoreExecutionSize = ( xn->XmlGetChildValueByName("nCore!ExecutionSize", m) ) ? m : 1;

	return filset;
}

bool ImageFilter::CanLocalize(){ return true; }


IntPtr ImageFilter::GetSupportedFormats()
{
	xml_node ^xn, ^node = gcnew xml_node( nullptr, "", "", "" );

	node->children.Add( gcnew xml_node(node, "InputImageCount", "", "1" ) );
	node->children.Add( gcnew xml_node(node, "OutputImageCount", "", "1" ) );
	xn = gcnew xml_node( node, "InputImage0", "", "" );
	node->children.Add( xn );
	//                     PinName, width, height, stride, packed, interleaved, channelcount, pixelsize, data)
	CreateImageSetting(xn, "Input",     0,      0,      0,      1,           1,            0,         0,    0);
	xn->XmlGetChildByName( xn, "ChannelsBitAllocation" );
	SetChannelsBitAllocation(xn, 0, 8);

	xn = gcnew xml_node( node, "OutputImage0", "", "" );
	node->children.Add( xn );
	//                     PinName, width, height, stride, packed, interleaved, channelcount, pixelsize, data)
	CreateImageSetting(xn, "Output",     0,      0,      0,      1,           1,            0,         0,    0);
	xn->XmlGetChildByName( xn, "ChannelsBitAllocation" );
	SetChannelsBitAllocation(xn, 0, 8);
	/*
	width,height = 2 --> width, height must be power of 2... = 8 --> multiply of 8
	packed, interleaved = 0, 1 (true, false)
	channelcount = 0 --> any positive number
	channelcount > 0 --> the value itself
	*/

	return (IntPtr)(void*)node->XML_Tree_Marshal();
}



// ***********************************************************************************************
// DO NOT MODIFY OR MOVE TO ELSEWHERE THESE HELPER FUNCTIONS
// THESE ARE HERE BECAUSE OF FilterCore IMPLEMENTED IN THIS FILE!!!

bool ImageFilter::FilterCore(filtersettings* f, int sx, int sy, int sw, int sh)
{
	return (::FilterCore(f, sx, sy, sw, sh, nullptr));
}

bool ImageFilter::ExecuteFilter(Byte* fs)
{ 
	bool b = false;
	if (!fs ) return false;
	xml_node^ xn = gcnew xml_node( fs );
	filtersettings* filset = XmlToMyStruct( xn, false );
	delete xn;
	if (filset)
	{
		*(filset->processing_ptr) = 0;
		*(filset->processing_size) = filset->width * filset->height;

		int cpucores = System::Environment::ProcessorCount;

		if (!filset->nCoreEnabled || cpucores == 1)
		{
			b = ::FilterCore(filset, 0, 0, 0, 0, nullptr); // FASTCALL WITH 0,0,0,0 PARAMETERS
		}
		else
		{
			filter_core_func f = (::FilterCore);

			if (FILTER_TYPE == IMAGEFILTER_PIXELWISE)
			{
				filset->nCoreExecutionSize = 1; // SAFETY CONDITIONING
				b = ImageFilterBase::ExecuteFilterBlock(filset, f);
			}
			else if (FILTER_TYPE == IMAGEFILTER_BOX)
			{
				b = ImageFilterBase::ExecuteFilterBox(filset, f);
			}
			else if (FILTER_TYPE == IMAGEFILTER_BLOCK)
			{
				b = ImageFilterBase::ExecuteFilterBlock(filset, f);
			}
			else // if (FILTER_TYPE == IMAGEFILTER_RECURSIVE) AND OTHERS
			{
				// NO AUTOMATIC MULTITHREAD SUPPORT
				b = ::FilterCore(filset, 0, 0, 0, 0, nullptr); // FASTCALL WITH 0,0,0,0 PARAMETERS
			}
		}
		delete filset;
		*(filset->processing_ptr) = *(filset->processing_size);
		*(filset->nCoreKillSignal) = 0; // RESET SIGNAL BIT IN THE MAIN PROGRAM
	}
	return b;
}
